/**
 * File:
 *  modbusMaster.c
 *
 * Notes:
 *  This file contains the implementation for a serial RTU modbus master.
 *
 * Functions:
 *  decodePacket      Decodes a packet received from modbus slave
 *  modifyBlock       Modifies either a coils or holding registers block
 *  nextSchedule      Finds the next block to poll
 *  scheduleRequest   Adds an I/O block to the polling schedule
 *  serviceSchedule   Services the polling schedule in round robin fashion
 *  unpackBits        Unpacks bits from a message into coils or status inputs
 *  unpackWords       Unpacks words from a message into input or holding regs.
 *
 * History:
 *  14/07/2012 Written by Simon Platten
 *  05/08/2012 Added unpackWords function
 *  31/08/2012 Added functions:
 *                   modifySingleCoil
 *                   modifyMSingleHoldingRegister
 *                   modifyMultiple
 *
 */
#include <stdarg.h>
#include <built_in.h>

#include "modbusMaster.h"

// Pointer to the polling schedule linked list
static volatile modbusPollScheduleDef* pPollSchedule = NULL;
static volatile modbusPollScheduleDef* pCurrPoll = NULL;
static volatile modbusPollScheduleDef* pLast;
static volatile boolean blnBusy = FALSE;
static volatile byte bytRetries = 0;
static volatile uint uintPacketCtr;

static volatile mbBlockStatus eCurrent, eLastStatus;
/**
 * Function:
 *  nextSchedule
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  0 if block found to poll, -1 if not
 */
static int nextSchedule(void) {
    if ( pCurrPoll != NULL ) {
      eLastStatus = pCurrPoll->eStatus;
// Make sure the current poll is ready for next time around
      pCurrPoll->eStatus = BLOCK_STATUS_READY;
    }
// Reset the retry counter
    bytRetries = 0;
// Find next item to poll, if current was already polled or the block is
// suspended or repeated maximum times
    pLast = pCurrPoll;
    while( TRUE ) {
      if ( pCurrPoll == NULL || pCurrPoll->pNext == NULL ) {
        pCurrPoll = pPollSchedule;

        if ( pLast == NULL ) {
          pLast = pCurrPoll;
        }
      } else {
        pCurrPoll = pCurrPoll->pNext;

        if ( pCurrPoll == pLast ) {
// Couldn't find block to read, abort!
          pCurrPoll = NULL;
          return -1;
        }
      }
      eCurrent = pCurrPoll->eStatus;

      if ( pCurrPoll->eStatus == BLOCK_STATUS_READY ) {
// Found block ready to poll
        return 0;
      }
    }
}
/**
 * Function:
 *  modifySingleCoil
 *
 * Parameters:
 *  pSchedule, a pointer to the schedule to modify
 *
 * Returns:
 *  none
 */
static void modifySingleCoil(modbusPollScheduleDef* pSchedule) {
  if ( pCurrPoll->arybytPendingWrite[0] > 0 ) {
    arybytMbBuffer[bytMbIndex++] = 0xff;
    arybytMbBuffer[bytMbIndex++] = 0x00;
  } else {
    arybytMbBuffer[bytMbIndex++] = 0x00;
    arybytMbBuffer[bytMbIndex++] = 0x00;
  }
}
/**
 * Function:
 *  modifySingleHoldingRegister
 *
 * Parameters:
 *  pSchedule, a pointer to the schedule to modify
 *
 * Returns:
 *  none
 */
static void modifySingleHoldingRegister(modbusPollScheduleDef* pSchedule) {
  arybytMbBuffer[bytMbIndex++] = pCurrPoll->arybytPendingWrite[1];
  arybytMbBuffer[bytMbIndex++] = pCurrPoll->arybytPendingWrite[0];
}
/**
 * Function:
 *  modifyMultiple
 *
 * Parameters:
 *  pSchedule, a pointer to the schedule to modify
 *
 * Returns:
 *  none
 */
static void modifyMultiple(modbusPollScheduleDef* pSchedule) {
  byte bytCount, bytIndex;
  arybytMbBuffer[bytMbIndex++] = Hi(pCurrPoll->uintItemsToModify);
  arybytMbBuffer[bytMbIndex++] = Lo(pCurrPoll->uintItemsToModify);
  if ( pCurrPoll->pBlock->eType == HOLDING_REGISTERS ) {
    bytCount = pCurrPoll->uintItemsToModify * 2;
  } else {
    bytCount = (pCurrPoll->uintItemsToModify / 8) +
               (pCurrPoll->uintItemsToModify % 8) ? 1 : 0;
  }
  arybytMbBuffer[bytMbIndex++] = bytCount;
  for( bytIndex=0; bytIndex<bytCount; bytIndex++ ) {
    arybytMbBuffer[bytMbIndex++] = pCurrPoll->arybytPendingWrite[bytIndex];
  }
}
/**
 * Function:
 *  decodePacket
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
#pragma funcall decodePacket dummy

void decodePacket(void) {
  if ( TMR2IF_bit ) {
    if ( uintPacketCtr < uintPacketTimeout ) {
      uintPacketCtr++;
    } else {
// Timeout! Try last message again
      if ( pCurrPoll
        && pCurrPoll->eStatus == BLOCK_STATUS_WAITING_FOR_RESPONSE ) {
// Increment the number of retries
        if ( ++bytRetries >= MAX_RETRIES ) {
// All attempts have failed, try next block
          nextSchedule();
        } else {
          pCurrPoll->eStatus = BLOCK_STATUS_RETRY;
        }
// Stop timer2
      TMR2ON_bit = 0;
      TMR2IE_bit = 0;
      blnBusy = FALSE;
      }
    }
    TMR2IF_bit = 0;
    return;
  }
  if ( TMR0IF_bit ) {
    bytMbRxphase.B0 = 1;
    TMR0ON_bit = 0;
// Clear the interrupt mask
    TMR0IF_bit = 0;

    if ( bytMbRxphase.B7 ) {
// Last byte sent, restart reception & set silent interval
      restartRx();
// Reset packet timeout counter & start counting
      uintPacketCtr = 0;
      TMR2ON_bit =  1;
      TMR2IE_bit = 1;
      return;
    }
  }
  if( RCIF_bit ) {
    byte bytTemp;
// Overrun or framing error?
    if ( OERR_bit || FERR_bit ) {
      bytTemp = RCREG;
      bytTemp = RCREG;
// Clear errors
      CREN_bit = 0;
      CREN_bit = 1;
      bytMbRxphase.B0 = 0;
      bytMbRxphase.B4 = 1;
      startTimeout(); // 4.5 chars
      return;
    }
// May need to store RX9D_bit here for parity check!
// Read received byte
    bytTemp = RCREG;
// Silent interval or inter-char interval passed?
// If passed, is there enough space in buffer?
    if ( bytMbRxphase == 0 
      || bytMbRxphase.B2 == 1 
      || bytMbRxphase.B4 == 1
      || bytMbIndex >= MAX_PACKET_LENGTH ) {
// No, wait for end of current message
      bytMbRxphase.B0 = 0;
      bytMbRxphase.B4 = 1;
      startTimeout();  // 4.5 chars
      return;
    }
    if ( bytMbIndex == 0) {
// Address received
      bytMbRxphase = 2;
    }
    if ( bytMbRxphase <= 2 ) {
// Store the data in buffer
      arybytMbBuffer[bytMbIndex++] = bytTemp;
    }
// Set inter-char interval
    startTimeout();  // 2.5 chars
  } else if ( bytMbRxphase.B7 ) {
    if ( TXIF_bit && TXIE_bit ) {
      if ( pCurrPoll && bytMbIndex < pCurrPoll->uintPL ) {
        TXREG = arybytMbBuffer[bytMbIndex++];
      } else {
// Wait for last byte to be transmitted
        startTimeout();  // 1 char
// Disable Tx interrupts
        TXIE_bit = 0;
      }
      return;
    }
  }
  if ( bytMbRxphase == 3 ) {
// Inter-char interval passed, check inter-frame timeout remainder
    bytMbRxphase = 4;
    startTimeout(); // 2 chars
  }
  if ( bytMbRxphase != 5 ) {
// Nothing more to do
    return;
  }
// Message received - long enough, right address?
  if ( bytMbIndex > 3 ) {
    if ( pCurrPoll 
      && arybytMbBuffer[0] == pCurrPoll->pBlock->bytSlaveAddress
      && (arybytMbBuffer[1] & 0x7f) == pCurrPoll->bytLastFC ) {
// No exception, calculate the packet CRC
      bytMbIndex -= 2;
      uintCRC = calcCRC();
// Compare the calculate checksum with the received checksum
      if ( Lo(uintCRC) == arybytMbBuffer[bytMbIndex] &&
           Hi(uintCRC) == arybytMbBuffer[bytMbIndex + 1] ) {
// CRC ok
        if ( (arybytMbBuffer[1] & EXCEPTION_FLAG) != EXCEPTION_FLAG
          &&  bytMbIndex >= (pCurrPoll->uintERL - 2) ) {
          switch( pCurrPoll->bytLastFC ) {
          case READ_COILS:
          case READ_STATUS_INPUTS:
            unpackBits(&arybytMbBuffer[2]);
            break;
          case READ_HOLDING_REGISTERS:
          case READ_INPUT_REGISTERS:
//            unpackWords(&arybytMbBuffer[2]);
            break;
          }
        }
// Next schedule
        nextSchedule();
// Valid message
        bytMbRxphase.B6 = 1;
        blnBusy = FALSE;
      }
    }
  }
// Message received
  bytMbRxphase.B5 = 1;
// If wrong message, invoke Timer2 interrupt with forced timeout
  if ( !bytMbRxphase.B6 ) {
// No longer waiting for response
    uintPacketCtr = uintPacketTimeout;
    TMR2IF_bit = 1;
  }
}
/**
 * Function:
 *  modifyBlock
 *
 * Parameters:
 *  pSchedule, a pointer to the schedule to modify
 *  uintAddress, the first address in the block to modify, base 1
 *  uintTotal, the total number of items to write
 *  pData, a pointer to the data to write to the block
 *
 * Returns:
 *  TRUE if ok, FALSE if not
 */
boolean modifyBlock(modbusPollScheduleDef* pSchedule, 
                    uint uintAddress, uint uintTotal, byte* pData) {
// Initialise variable argument list
  if ( pSchedule == NULL
    || pSchedule->pBlock == NULL
    || uintAddress == 0
    || uintTotal == 0
    || pData == NULL
    || uintAddress < pSchedule->pBlock->uintAddress
    || (uintAddress + uintTotal - 1) >
       (pSchedule->pBlock->uintAddress + pSchedule->pBlock->uintTotal - 1)
    || !(pSchedule->pBlock->eType == COILS 
      || pSchedule->pBlock->eType == HOLDING_REGISTERS) ) {
    return FALSE;
  }
// Save the start address and initialise the items to modify counter
  pSchedule->uintStartAddress = uintAddress - 1;
  pSchedule->uintItemsToModify = uintTotal;
// Transfer the bytes to write into the pending buffer
  if ( pSchedule->pBlock->eType == COILS ) {
    memcpy(pSchedule->arybytPendingWrite, pData, 
           (uintTotal / 8) + ((uintTotal % 8) ? 1 : 0));
  } else {
    memcpy(pSchedule->arybytPendingWrite, pData, (uintTotal * 2));
  }
  if ( pSchedule->pBlock->eType == COILS ) {
    if ( pSchedule->uintItemsToModify == 1 ) {
      pSchedule->pModify = modifySingleCoil;
    } else {
      pSchedule->pModify = modifyMultiple;
    }
  } else {
    if ( pSchedule->uintItemsToModify == 1 ) {
      pSchedule->pModify = modifySingleHoldingRegister;
    } else {
      pSchedule->pModify = modifyMultiple;
    }
  }
  return TRUE;
}
/**
 * Function:
 *  scheduleRequest
 *
 * Parameters:
 *  pBlock, a pointer to the I/O block to schedule a poll for
 *  pSchedule, a pointer to the schedule node to set-up
 *
 * Returns:
 *  TRUE if block added, FALSE if not
 */
boolean scheduleRequest(modbusBlockDef* pBlock, 
                        modbusPollScheduleDef* pSchedule) {
// Validate parameters
  if ( pBlock == NULL
    || pBlock->eType == INVALID_MB_TYPE
    || pBlock->uintAddress == 0
    || pBlock->uintTotal == 0
    || pBlock->paryData == NULL ) {
    return FALSE;
  }
  if ( pPollSchedule == NULL ) {
// This is the first block
    pPollSchedule = pSchedule;
// Set-up Timer 2 to be used as the response packet timeout
// Prescaler=1:16;
// TMR2 PostScaler=1:10;
// PR2=250 - Period = 5.00 ms
    T2CON = 0b01001010;
    PR2 = 249;          // PR2 (Timer2 Match value)
  } else {
// Make sure there are no blocks with the same address and type
    modbusPollScheduleDef* pNode;
    for( pNode=pPollSchedule; pNode->pNext!=NULL; pNode=pNode->pNext ) {
      if ( pNode->pBlock == pBlock ) {
// A block already exists using this I/O block, abort
        return FALSE;
      }
    }
// Append this block to the linked list
    pNode->pNext = pSchedule;
  }
  pSchedule->pBlock  = pBlock;
  pSchedule->pNext   = NULL;
  pSchedule->pModify = NULL;
  pSchedule->uintItemsToModify = 0;
  pSchedule->eStatus = BLOCK_STATUS_READY;
  return TRUE;
}
/**
 * Function:
 *  serviceSchedule
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
void serviceSchedule(void) {
  uint uintAddress;
#if ( MODBUS_INTERPOLL_DELAY > 0 )
  Delay_ms(MODBUS_INTERPOLL_DELAY);
#endif
  if ( blnBusy == TRUE ) {
// Do nothing while busy, this means a request is already pending response
    return;
  }
  if ( pCurrPoll == NULL ) {
// Set-up current item to poll
    pCurrPoll = pPollSchedule;
  }
  if ( !(pCurrPoll->eStatus == BLOCK_STATUS_READY
      || pCurrPoll->eStatus == BLOCK_STATUS_RETRY) ) {
    return;
  }
// Change the schedule status
  pCurrPoll->eStatus = BLOCK_STATUS_WAITING_FOR_RESPONSE;
// Create request message
  if ( pCurrPoll->pBlock->eType == COILS ) {
    if ( pCurrPoll->pModify == NULL ) {
      pCurrPoll->bytLastFC = READ_COILS;
    } else if ( pCurrPoll->uintItemsToModify == 1 ) {
      pCurrPoll->bytLastFC = FORCE_SINGLE_COIL;
    } else {
      pCurrPoll->bytLastFC = FORCE_MULTIPLE_COILS;
    }
  } else if ( pCurrPoll->pBlock->eType == STATUS_INPUTS ) {
    pCurrPoll->bytLastFC = READ_STATUS_INPUTS;
  } else if ( pCurrPoll->pBlock->eType == HOLDING_REGISTERS ) {
    if ( pCurrPoll->pModify == NULL ) {
      pCurrPoll->bytLastFC = READ_HOLDING_REGISTERS;
    } else if ( pCurrPoll->uintItemsToModify == 1 ) {
      pCurrPoll->bytLastFC = PRESET_SINGLE_REGISTER;
    } else {
      pCurrPoll->bytLastFC = PRESET_MULTIPLE_REGISTERS;
    }
  } else if ( pCurrPoll->pBlock->eType == INPUT_REGISTERS ) {
    pCurrPoll->bytLastFC = READ_INPUT_REGISTERS;
  } else {
// Invalid type, abort!
    return;
  }
// Calculate the expected response length
  if ( pCurrPoll->pBlock->eType == COILS
    || pCurrPoll->pBlock->eType == STATUS_INPUTS ) {
// Length = Slave Address + Function Code + Byte Count + I/O + CRC
    pCurrPoll->uintERL = 5 + 
                         ((pCurrPoll->pBlock->uintTotal / 8) +
                          (((pCurrPoll->pBlock->uintTotal % 8) > 0) ? 1 : 0));
  } else if ( pCurrPoll->pBlock->eType == HOLDING_REGISTERS
           || pCurrPoll->pBlock->eType == INPUT_REGISTERS ) {
// Length = Slave Address + Function Code + Byte Count + I/O + CRC
    pCurrPoll->uintERL = 5 + (pCurrPoll->pBlock->uintTotal * 2);
  }
  bytMbIndex = 0;
// Slave address
  arybytMbBuffer[bytMbIndex++] = pCurrPoll->pBlock->bytSlaveAddress;
// Function code
  arybytMbBuffer[bytMbIndex++] = pCurrPoll->bytLastFC;
  
  if ( pCurrPoll->pModify != NULL ) {
// Populate the start address, common to all write functions
    arybytMbBuffer[bytMbIndex++] = Hi(pCurrPoll->uintStartAddress);
    arybytMbBuffer[bytMbIndex++] = Lo(pCurrPoll->uintStartAddress);
// Call the specific modify routine to build the message
    (*pCurrPoll->pModify)(pCurrPoll);
// Clear out the modify function pointer
    pCurrPoll->pModify = NULL;
  } else {
// Start Address
    uintAddress = pCurrPoll->pBlock->uintAddress - 1;
    arybytMbBuffer[bytMbIndex++] = Hi(uintAddress);
    arybytMbBuffer[bytMbIndex++] = Lo(uintAddress);
// Number of items
    arybytMbBuffer[bytMbIndex++] = Hi(pCurrPoll->pBlock->uintTotal);
    arybytMbBuffer[bytMbIndex++] = Lo(pCurrPoll->pBlock->uintTotal);
  }
// Calculate CRC for response
  uintCRC = calcCRC();
  arybytMbBuffer[bytMbIndex++] = Lo(uintCRC);
  arybytMbBuffer[bytMbIndex++] = Hi(uintCRC);
// Store length and set index
  pCurrPoll->uintPL = bytMbIndex;
// Wait for end of last trasmission or silent interval
  while( bytMbRxphase.B0 == 0 ) ;
// Stop receiving
  CREN_bit = 0;
  RCIE_bit = 0;
// Enable transmission, set direction
  TXEN_bit = 1;
  Tx_dir = 1;
// Mark start of transmission
  bytMbRxphase = 128;
// Send address
  TXREG = arybytMbBuffer[0];
// Send function code to fill Tx buffer and avoid instant return to ISR
  while( !TXIF_bit ){}
  TXREG = arybytMbBuffer[1];
// Indicate we are busy and waiting for response
  blnBusy = TRUE;
// Enable Tx interrupts for remainder of response to follow
  bytMbIndex = 2;
  TXIE_bit = 1;
}
/**
 * Function:
 *  unpackBits
 *
 * Parameters:
 *  pBuffer, a pointer to the message data starting from byte count
 *
 * Returns:
 *  none
 */
void unpackBits(byte* pBuffer) {
  byte* pbytData, bytBit, bitMask;
  byte bytByteCount;
// Get the number of databytes to decode
  bytByteCount = *pBuffer;
  pBuffer++;
  pCurrBlock = pCurrPoll->pBlock;
  pbytData = (byte*)pCurrBlock->paryData;
  while( bytByteCount-- ) {
    for( bytBit=0; bytBit<8; bytBit++ ) {
      bitMask = 1 << bytBit;

      if ( (*pBuffer & bitMask) == bitMask ) {
        *pbytData |= bitMask;
      } else {
        *pbytData &= (0xff ^ bitMask);
      }
    }
    pbytData++;
    pBuffer++;
  }
// If there is a callback associated with the I/O block then call it now
  if ( pCurrBlock->pCallback != NULL ) {
    (*pCurrBlock->pCallback)();
  }
}
/**
 * Function:
 *  unpackBits
 *
 * Parameters:
 *  pBuffer, a pointer to the message data starting from byte count
 *
 * Returns:
 *  none
 */
void unpackWords(byte* pBuffer) {
  byte bytByteCount;
  uint* puintData;
// Get the number of databytes to decode
  bytByteCount = *pBuffer;
  pBuffer++;
  pCurrBlock = pCurrPoll->pBlock;
  puintData = (uint*)pCurrBlock->paryData;
  while( bytByteCount-- ) {
    Hi(puintData) = *pBuffer;
    pBuffer++;
    Lo(puintData) = *pBuffer;
    pBuffer++;
    puintData++;
  }
// If there is a callback associated with the I/O block then call it now
  if ( pCurrBlock->pCallback != NULL ) {
    (*pCurrBlock->pCallback)();
  }
}